;(load "unification.scm")
;(load "rule.scm")
;(load "sets.scm")





; (module-export nd-queue)
; (module-export nd-queue-init)

; (module-export graph-make)
; (module-export state-clean)
; (module-export couplerulesigma-apply )
; (module-export rules-firstrule-get-future-couplerulesigmatoapply)
; (module-export couplerulesigma-getrule)
; (module-export couplerulesigma-getsigma)
; (module-export couplerulesigma-getinstanciedconditions)


;this is the non-deterministic queue. It contains the possible states to handle if the current execution fails
(define nd-queue '())

;init the non-deterministic queue to the empty queue
(define (nd-queue-init) (set! nd-queue '()))

;add a state to the non-deterministic queue to come back later if the current execution fails
(define (nd-queue-push state)
  (set! nd-queue (cons state nd-queue)))

;gives a state that has to treat and remove it from the non-deteministic queue (in case of failing)
(define (nd-queue-pop)
  (if (null? nd-queue)
      #f
      (let ((currentstate (car nd-queue))
            (tail (cdr nd-queue)))
        (begin (set! nd-queue tail)
               currentstate))))

;gives the next state to deal with of the non-deterministic queue
(define (nd-queue-top)
  (car nd-queue))

;make a graph with the set of terms S
(define (graph-make S)
  `(graph ,S))

;used for printing debugging stuff on the command line
(define (chatterboxsays m) '())
;(define (chatterboxsays m) (begin (display m) (newline)))

(define (progressbar) (display "-"))
(define (progressbar-ruleapplied) (display "+"))

;eval all the unquoted expression in a list
;(listeval '(2 ,(+ 1 2))) returns '(2 3)
(define (listeval L)
  (if (list? L)
      (if (null? L)
          '()
          (if (equal? (car L) 'unquote)
              (eval (cadr L))
              (cons (listeval (car L)) (listeval (cdr L)))))
      L))


;returns a new machine state where whatweadd has been added
;~O(1)
(define (kb-add state whatweadd)
  (if (equal? whatweadd '((halt)))
      ;we close a tableau branch if we find a halt
      (begin (display "!") (newline) (nd-queue-pop))
	  ;otherwise add!
      (machinestate-elementsadd state whatweadd )))


;~O(1)
; state = the current branch
; sigma = the substitution
; rule = the considered rule
; addpatterns = the list of patterns to add (generally the size of the list is 1)
;returns : a computation where variables of addpatterns are instanciated and expressions are computed
(define (kb-chooseaddpatterns-add state sigma rule)
	(let ((addpatterns (rule-addspatterns-get rule)))
  (begin
   ;inform the user in case of a non-deterministic rule
    (if (> (set-card addpatterns) 1)
        (begin (display "<") (newline) )
        '()) ;else do nothing
		
		
    (let* (
		(subssigma (lambda (p) (substituteBySub1InTerm2 `((set (quote ,(machinestate-set-get state))))
                                                                                         (substituteBySub1InTerm2 
                                                                                          sigma
                                                                                          p))))
		(sigmafromletpart (substitution-values-eval (subssigma (rule-let-get rule))))
	
	(prepareaddpatterns (lambda (p)  (listeval (free-variables-instanciate (substituteBySub1InTerm2 sigmafromletpart (subssigma p))))))) 
      (if (null? (cdr addpatterns))
          ;determistic
          (kb-add state (prepareaddpatterns (car addpatterns)))
          ;non-deterministic
          (begin
            (nd-queue-push 
             (kb-add (machinestate-copy state) (prepareaddpatterns (cadr addpatterns))))
            (kb-add state (prepareaddpatterns (car addpatterns)))))))))


;returns a substitution sigma from sigmaset such that no subset of addpatterns is in the state
;if no, returns #f
;addpatterns = set of patterns to add to the state
;sigmaset = set of substitutions
(define (rule-application-substitution-where-addpatterns-not-in-get state sigmaset addpatterns)	  
  (if (null? sigmaset)
      #f
      (let* ((sigma (car sigmaset))
             (set (machinestate-set-get state))
             (sigmaaddpatterns (substituteBySub1InTerm2 `((set (quote ,set)))
                                                        (substituteBySub1InTerm2 sigma addpatterns))))
        
        (if (unifiableOnePatternsSetsOfSet1WithElementsSet2?
             (listeval sigmaaddpatterns)
             (machinestate-set-get state))
            ;if the pattern to add is already in kb, we continue to look in (cdr sigmaset)
            (rule-application-substitution-where-addpatterns-not-in-get state (cdr sigmaset) addpatterns)
            sigma
            ))))


;check that the condition expressed in suchthattoexecute is satisfied
;sigma = substitution for terms in suchthattoexecute
;set = the set of ALL terms in the state
;example: (sigmaok? '((A 1) (B 2)) '(+ A B set) 5)
; returns 8
(define (sigmaok? sigma suchthattoexecute set)
  (let ((programtotest (substituteBySub1InTerm2 `((set (quote ,set))) (substituteBySub1InTerm2 (substitution-values-quote sigma) suchthattoexecute)) ))
      (eval programtotest)))



;remove from sigmasetconditionpos all the substitutions that do not satisfy the condition expressed in suchthattoexecute
;returns this filtered set of substitutions
;sigmasetconditionpos = set of substitutions
;suchthattoexecute = a scheme expression
;set = the set of ALL terms in the state
(define (sigmasetfilter sigmasetconditionpos suchthattoexecute set)
  (if (null? sigmasetconditionpos)
      '()
      (if (sigmaok? (car sigmasetconditionpos) suchthattoexecute set)
          (cons (car sigmasetconditionpos) (sigmasetfilter (cdr sigmasetconditionpos) suchthattoexecute set))
          (sigmasetfilter (cdr sigmasetconditionpos) suchthattoexecute set)
          )
      ))


;return the set sigmaset where we have removed the substitution such that the doesnotcontain is included in the set set (that represent the set of terms of the current state)
(define (sigmasetdelete sigmaset doesnotcontain set)
  (if (null? sigmaset)
      '()
      (let ( (sigma (car sigmaset)) )
        (if (set-include? (substituteBySub1InTerm2 sigma doesnotcontain) set)
            (sigmasetdelete (cdr sigmaset) doesnotcontain set)
            (cons sigma (sigmasetdelete (cdr sigmaset) doesnotcontain set))))))


;returns a substitution sigma making the rule applicable
;the substitution should unify the term newterm with an element of the condition part
;returns #f (false) if it is not applicable
(define (rule-application-substitution-get rule state newterm)
  (let* (;condition part of the rule
		 (conditionpos (rule-positiveconditionspatterns-get rule))
		 
		 ;the set of substitutions that mathes newterm and the condition part
         (sigmasetfromnewterm (unifyOnePatternOfSet1WithElement2 conditionpos newterm))
		 
		 ;complete the set of substitutions so that all the variables of the condition part are instanciated and matches with elemnets from the state
         (sigmasetconditionpos 
          (unifyPatternsSet1WithElementsOfSet2AccordingToSubsSet3 
           conditionpos
           (machinestate-set-get state)
           sigmasetfromnewterm))
		   
		  ;filter from this set of substitutions all the substitutions that makes the "such that" condition true
         (sigmasetwithsuchthat (sigmasetfilter sigmasetconditionpos (rule-suchthat-get rule) (machinestate-set-get state)))
		 
		 ;takes in account the "doesnotcontain part"
         (sigmaset 
          (if (rule-doesnotcontain-get rule)
              (sigmasetdelete sigmasetwithsuchthat (rule-doesnotcontain-get rule) (machinestate-set-get state))
              sigmasetwithsuchthat))
         )
    (if (set-empty? sigmaset)
        (begin
          (chatterboxsays `(":( rule doesn't match :) : " ,rule))
          ;(machinestate-compterendu-display state) 
          ;(progressbar)
          #f)
        (begin
          (chatterboxsays `(":---)))) rule matches : " ,rule ))
          ;(progressbar-ruleapplied)
          ;(machinestate-compterendu-display state)
		  
		  ;if the rule merges nodes etc. returns the corresponding sigmaset
			(if (rule-merge-get rule)
			   ; return the first substitution in sigmaset
				 (car sigmaset)
		  ;else remove from sigmaset the substitutions that will not add new patterns
			  (rule-application-substitution-where-addpatterns-not-in-get 
			   state 
			   sigmaset
			   (rule-addspatterns-get rule)))
          )
        )))




;rename recursively all occurences in S of e2 by e1
(define (replaceinS1e23bye12 S e1 e2)
  (if (equal? S e2)
      e1
      (if (list? S)
          (if (null? S)
              '()
              (cons (replaceinS1e23bye12 (car S) e1 e2) (replaceinS1e23bye12 (cdr S) e1 e2)))
          (if (vector? S)
              (let ((f (lambda (l) (replaceinS1e23bye12 l e1 e2))))
                (vector-map S f))
              S))))


;fuse the term e1 and e2 in S and makes a new machine state from it
(define (fusemerge S e1 e2)
  (machinestate-nodoublons (replaceinS1e23bye12 S e1 e2)))


; apply the rule rule with the substitution sigma for variables in the condition of the rule
;and returns the new state
(define (rule-apply rule sigma state)
  ;statewithadd corresponds to the state state with what we have added
  (let* ((statewithadd (kb-chooseaddpatterns-add state sigma rule))
         (mergevars (rule-merge-get rule) )
         (statewithmerge
          (if mergevars
              (begin
                (display "merge ")
                (display (substitution-variable-valeur-get sigma (car mergevars)))
                (display " and ")
                (display (substitution-variable-valeur-get sigma (cadr mergevars)))
                (display ".")
                (newline)
                (fusemerge statewithadd (substitution-variable-valeur-get sigma (car mergevars))
                           (substitution-variable-valeur-get sigma (cadr mergevars))))
              statewithadd)))
    (begin
      ;            (display "we apply rule ")
      ;            (display rule)
      ;            (display " with substitution ")
      ;            (display sigma)
      ;            (display " launched by the event ")
      ;            (display newterm)
      ;            (display ". ")
      statewithmerge)))


;find a substitution with the help of newterms
;returns the substitution
;of #f if no substitution
(define (rule-apply-find-newterm rule state newterms)
  (if (null? newterms)
      #f
      (let ((sigma (rule-application-substitution-get rule state (car newterms))))
        (if sigma
            sigma
            (begin
              (rule-apply-find-newterm rule 
                                       state
                                       (cdr newterms)))))))



;try to get a good substitution to apply rule rule on state
(define (rule-apply-try rule state)
  (rule-apply-find-newterm rule state (machinestate-events-forrule-get state rule)))


(define (rule-get-couple-cleanedstate-sigma-newterm rule state newterms)
  (if (null? newterms)
      (list state #f)
      (let ( (sigma (rule-application-substitution-get rule state (car newterms)))
             (cleaned-state (machinestate-event-delete state rule (car newterms))))
        (if sigma
            (list cleaned-state sigma)
            (begin
              (rule-get-couple-cleanedstate-sigma-newterm rule 
                                                          cleaned-state
                                                          (cdr newterms)))))))

(define (rule-get-couple-cleanedstate-sigma rule state)
  (rule-get-couple-cleanedstate-sigma-newterm rule state (machinestate-events-forrule-get state rule)))

  
 ;;the aim is to remove ALL the newterms that are useless
(define (state-clean-rule-newterms rule newterms state)
  (if (null? newterms)
      state
      (if (rule-application-substitution-get rule state (car newterms))
          (state-clean-rule-newterms rule (cdr newterms) state)
          ;state
          (state-clean-rule-newterms rule (cdr newterms) (machinestate-event-delete state rule (car newterms))))))

		  
;we remove all what is unecessary concerning the rule in the state
(define (state-clean-rule rule state)
  (state-clean-rule-newterms rule (machinestate-events-forrule-get state rule) state))

  
;;MAIN FUNCTION (up to now)  
;remove the useless event
; (define (state-clean rules state)
; (if (null? rules)
; state
; (state-clean (cdr rules) (state-clean-rule (car rules) state))))
(define (state-clean rules state)
  (if (null? rules)
      state
      (let ((sigma (rule-apply-try (car rules) state)))
        (if sigma
            (state-clean-rule (car rules) state)
            (state-clean (cdr rules) (state-clean-rule (car rules) state))))))






;gives the positive conditions part of the rule, instancied with the substitution
(define (couplerulesigma-getinstanciedconditions C) 
  (let ((rule (couplerulesigma-getrule C))
        (sigma (couplerulesigma-getsigma C)) )
    (substituteBySub1InTerm2 sigma (rule-positiveconditionspatterns-get rule))))

;accessor for having the substitution sigma of the rule
(define (couplerulesigma-getrule C)	(car C))
(define (couplerulesigma-getsigma C) (cadr C))





;MAIN FUNCTION TO COMPUTE THE NEXT RULE AND PATTERN TO APPLY
;give me the next (rule, sigma) to apply on state, where rule is a tableau rule and sigma is a substitution
; this function ranges over the rules and takes in account the first rule that is applicable
(define (rules-firstrule-get-future-couplerulesigmatoapply rules state)
  (if (null? rules)
      #f
      (let ((sigma (rule-apply-try (car rules) state)))
        (if sigma
            `(,(car rules) ,sigma)
            (rules-firstrule-get-future-couplerulesigmatoapply (cdr rules) state)))))
			

;MAIN FONCTION TO APPLY THE CURRENT RULE
;This function apply the rule with the substitution sigma given in C. (it also clean the state before applying the rule)
;state : all the tableau + extra information in order to apply rules
;C : a couple (rule, sigma) where rule is tableau rule and sigma is a substitution for the variables of the rule condition
(define (couplerulesigma-apply C state)
  (rule-apply (couplerulesigma-getrule C) (couplerulesigma-getsigma C) (state-clean-rule (couplerulesigma-getrule C) state)))




